/*
 * Copyright (c) 2011 J. Lewis Muir <jlmuir@imca.aps.anl.gov>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package ch.psi.imagej.cbf;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;

import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.Raster;
import java.io.IOException;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.List;
import java.util.Locale;

import javax.imageio.IIOException;
import javax.imageio.ImageReadParam;
import javax.imageio.ImageReader;
import javax.imageio.ImageWriteParam;
import javax.imageio.event.IIOReadProgressListener;
import javax.imageio.event.IIOReadWarningListener;
import javax.imageio.stream.ImageInputStream;

import org.junit.Test;

import ch.psi.imagej.cbf.Base64;
import ch.psi.imagej.cbf.CbfCompression;
import ch.psi.imagej.cbf.CbfElementType;
import ch.psi.imagej.cbf.CbfImageReader;
import ch.psi.imagej.cbf.CbfImageReaderSpi;
import ch.psi.imagej.cbf.CbfMetadata;

/**
 * I test <code>CbfImageReader</code>.
 */
public class CbfImageReaderTest extends TestUtils {
  /* Ensure OK to invoke getImageMetadata twice. */
  @Test
  public void testGetImageMetadata01() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    ImageInputStream stream = streamFromBytes(createImageWithoutData());
    reader.setInput(stream);
    reader.getImageMetadata(0);
    reader.getImageMetadata(0);
  }

  /* Ensure state is reset if new input source is set. */
  @Test
  public void testSetInput01() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    ImageInputStream stream = streamFromBytes(createImageWithoutData());
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals("thaumatin_0001", mdata.getDataBlockName());
    String text = createImageText().replaceFirst("data_thaumatin_0001", "data_thaumatin_0002");
    stream = streamFromBytes(createImageWithoutData(asciiBytes(text)));
    reader.setInput(stream);
    mdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals("thaumatin_0002", mdata.getDataBlockName());
  }

  @Test
  public void testReadMetadataIdentifier01() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    ImageInputStream stream = streamFromBytes(createImageWithoutData());
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    String expected = "###CBF: VERSION 1.5, CBFlib v0.7.8 - SLS/DECTRIS PILATUS detectors";
    assertEquals(expected, mdata.getIdentifier());
  }

  @Test
  public void testReadMetadataIdentifier02() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    String text = createImageText().replaceFirst(
      "###CBF: VERSION 1.5, CBFlib v0.7.8 - SLS/DECTRIS PILATUS detectors",
      "###CBF: VERSION 1.5\r\n# CBF file written by CBFlib v0.7.8");
    ImageInputStream stream = streamFromBytes(createImageWithoutData(asciiBytes(text)));
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    String expected = "###CBF: VERSION 1.5";
    assertEquals(expected, mdata.getIdentifier());
  }

  @Test
  public void testReadMetadataDataBlockName01() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    ImageInputStream stream = streamFromBytes(createImageWithoutData());
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    String expected = "thaumatin_0001";
    assertEquals(expected, mdata.getDataBlockName());
  }

  @Test
  public void testReadMetadataHeaderConvention01() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    ImageInputStream stream = streamFromBytes(createImageWithoutData());
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    String expected = "SLS_1.0";
    assertEquals(expected, mdata.getHeaderConvention());
  }

  @Test
  public void testReadMetadataHeaderConvention02() throws IOException, InterruptedException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    String text = createImageText().replaceFirst(
      "_array_data\\.header_convention \"SLS_1\\.0\"",
      "_array_data.header_convention \"SLS_1.1\"");
    ImageInputStream stream = streamFromBytes(createImageWithoutData(asciiBytes(text)));
    reader.setInput(stream);
    final List<String> warnings = new ArrayList<String>();
    reader.addIIOReadWarningListener(new IIOReadWarningListener() {
      @Override
      public void warningOccurred(ImageReader source, String warning) {
        warnings.add(warning);
      }
    });
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals("SLS_1.1", mdata.getHeaderConvention());
    assertEquals(1, warnings.size());
    assertEquals("Unrecognized header convention: SLS_1.1", warnings.get(0));
  }

  /* Header convention without quotes. */
  @Test
  public void testReadMetadataHeaderConvention03() throws IOException, InterruptedException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    String text = createImageText().replaceFirst(
      "_array_data\\.header_convention \"SLS_1\\.0\"",
      "_array_data.header_convention SLS_1.0");
    ImageInputStream stream = streamFromBytes(createImageWithoutData(asciiBytes(text)));
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals("SLS_1.0", mdata.getHeaderConvention());
  }

  /* Header convention not starting with "SLS_1.". */
  @Test(expected = IIOException.class)
  public void testReadMetadataHeaderConvention04() throws IOException, InterruptedException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    String text = createImageText().replaceFirst(
      "_array_data\\.header_convention \"SLS_1\\.0\"",
      "_array_data.header_convention \"SLS_2.0\"");
    ImageInputStream stream = streamFromBytes(createImageWithoutData(asciiBytes(text)));
    reader.setInput(stream);
    reader.getImageMetadata(0);
  }

  @Test
  public void testReadMetadataHeader01() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    ImageInputStream stream = streamFromBytes(createImageWithoutData());
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals("PILATUS 6M, 60-0103, IMCA-CAT", mdata.getHeaderDetector());
    Calendar c = Calendar.getInstance(Locale.US);
    c.set(Calendar.YEAR, 2010);
    c.set(Calendar.MONTH, 10);
    c.set(Calendar.DAY_OF_MONTH, 17);
    c.set(Calendar.HOUR_OF_DAY, 15);
    c.set(Calendar.MINUTE, 52);
    c.set(Calendar.SECOND, 30);
    c.set(Calendar.MILLISECOND, 834);
    assertEquals(c.getTime(), mdata.getHeaderDate());
    assertArrayEquals(new float[] { 172e-6f, 172e-6f }, mdata.getHeaderPixelSize(), 0.0f);
    assertEquals("Silicon", mdata.getHeaderSensorType());
    assertEquals((Float)0.000320f, mdata.getHeaderSensorThickness());
    assertEquals((Float)0.081033f, mdata.getHeaderExposureTime());
    assertEquals((Float)0.083333f, mdata.getHeaderExposurePeriod());
    assertEquals((Float)124.0e-09f, mdata.getHeaderTau());
    assertEquals((Integer)326741, mdata.getHeaderCountCutoff());
    assertEquals((Float)7439.0f, mdata.getHeaderThreshold());
    assertNull(mdata.getHeaderGainType());
    assertEquals((Float)9.900f, mdata.getHeaderGainVrf());
    assertEquals((Integer)2391, mdata.getHeaderNumExcludedPixels());
    assertEquals("badpix_mask.tif", mdata.getHeaderExcludedPixels());
    assertNull(mdata.getHeaderFlatField());
    assertNull(mdata.getHeaderTrim());
    assertEquals((Float)1.0000f, mdata.getHeaderWavelength());
    assertArrayEquals(new float[] { 0.0f, 0.0f }, mdata.getHeaderEnergyRange(), 0.0f);
    assertEquals((Float)0.94997f, mdata.getHeaderDetectorDistance());
    assertEquals((Float)0.0f, mdata.getHeaderDetectorVOffset());
    assertArrayEquals(new float[] { 1253.0f, 1288.0f }, mdata.getHeaderBeamXY(), 0.0f);
    assertEquals((Float)0.0f, mdata.getHeaderFlux());
    assertEquals((Float)0.0f, mdata.getHeaderFilterTransmission());
    assertEquals((Float)0.00f, mdata.getHeaderStartAngle());
    assertEquals((Float)1.0f, mdata.getHeaderAngleIncrement());
    assertEquals((Float)0.0007f, mdata.getHeaderDetector2Theta());
    assertEquals((Float)0.99f, mdata.getHeaderPolarization());
    assertEquals((Float)0.0f, mdata.getHeaderAlpha());
    assertEquals((Float)0.0f, mdata.getHeaderKappa());
    assertEquals((Float)0.0f, mdata.getHeaderPhi());
    assertEquals((Float)0.0f, mdata.getHeaderChi());
    assertEquals("X, CW", mdata.getHeaderOscillationAxis());
    assertEquals((Integer)1, mdata.getHeaderNumOscillations());
  }

  /* Two date formats. */
  @Test
  public void testReadMetadataHeader02() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    ImageInputStream stream = streamFromBytes(createImageWithoutData());
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    Calendar c = Calendar.getInstance(Locale.US);
    c.set(Calendar.YEAR, 2010);
    c.set(Calendar.MONTH, 10);
    c.set(Calendar.DAY_OF_MONTH, 17);
    c.set(Calendar.HOUR_OF_DAY, 15);
    c.set(Calendar.MINUTE, 52);
    c.set(Calendar.SECOND, 30);
    c.set(Calendar.MILLISECOND, 834);
    assertEquals(c.getTime(), mdata.getHeaderDate());
    String text = createImageText().replaceFirst(
      "# 2010-Nov-17T15:52:30.834\r\n",
      "# 2010/Nov/17 15:52:30.834\r\n");
    stream = streamFromBytes(createImageWithoutData(asciiBytes(text)));
    reader.setInput(stream);
    mdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals(c.getTime(), mdata.getHeaderDate());
  }

  /* "(nil)" excluded pixels and trim directory; not "(nil)" flat field. */
  @Test
  public void testReadMetadataHeader03() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    String text = createImageText().replaceFirst(
      "# Excluded_pixels: badpix_mask.tif\r\n",
      "# Excluded_pixels: (nil)\r\n");
    text = text.replaceFirst(
      "# Flat_field: \\(nil\\)\r\n",
      "# Flat_field: flat_field.tif\r\n");
    text = text.replaceFirst(
      "Trim_directory: \r\n",
      "Trim_directory: (nil)\r\n");
    ImageInputStream stream = streamFromBytes(createImageWithoutData(asciiBytes(text)));
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    assertNull(mdata.getHeaderExcludedPixels());
    assertEquals("flat_field.tif", mdata.getHeaderFlatField());
    assertNull(mdata.getHeaderTrim());
  }

  /* not "(nil)" trim directory. */
  @Test
  public void testReadMetadataHeader04() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    String text = createImageText().replaceFirst(
      "Trim_directory: \r\n",
      "Trim_directory: p6m0103_T8p0_vrf_m0p3_100323\r\n");
    ImageInputStream stream = streamFromBytes(createImageWithoutData(asciiBytes(text)));
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals("p6m0103_T8p0_vrf_m0p3_100323", mdata.getHeaderTrim());
  }

  @Test
  public void testReadMetadataMimeHeader01() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    ImageInputStream stream = streamFromBytes(createImageWithoutData());
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals(CbfCompression.BYTE_OFFSET, mdata.getDataCompression());
    assertEquals(6224001, mdata.getDataSize());
    assertEquals(1, mdata.getDataId());
    assertEquals(CbfElementType.SIGNED_32_BIT_INTEGER, mdata.getDataElementType());
    assertEquals(ByteOrder.LITTLE_ENDIAN, mdata.getDataElementByteOrder());
    assertArrayEquals(Base64.decode("Ti67wd6I6DPuGIsOnMalMw=="), mdata.getDataMd5());
    assertEquals(6224001, mdata.getDataNumElements());
    assertEquals(2463, mdata.getDataWidth());
    assertEquals(2527, mdata.getDataHeight());
    assertEquals(4095, mdata.getDataPadding());
  }

  @Test
  public void testReadMetadataMimeHeader02() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    String text = createImageText().replaceFirst(
      " conversions=\"x-CBF_BYTE_OFFSET\"",
      " conversions=\"x-CBF_NONE\"");
    ImageInputStream stream = streamFromBytes(createImageWithoutData(asciiBytes(text)));
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals(CbfCompression.NONE, mdata.getDataCompression());
  }

  /* Missing content type conversions. */
  @Test(expected = IIOException.class)
  public void testReadMetadataMimeHeader03() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    String text = createImageText();
    text = text.replaceFirst(
      "Content-Type: application/octet-stream;",
      "Content-Type: application/octet-stream");
    text = text.replaceFirst(
      " +conversions=\"x-CBF_BYTE_OFFSET\"\r\n",
      "");
    ImageInputStream stream = streamFromBytes(createImageWithoutData(asciiBytes(text)));
    reader.setInput(stream);
    reader.getImageMetadata(0);
  }

  @Test
  public void testReadMetadataMimeHeader04() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    String text = createImageText().replaceFirst(
      "X-Binary-Element-Type: \"signed 32-bit integer\"",
      "X-Binary-Element-Type: \"unsigned 32-bit integer\"");
    ImageInputStream stream = streamFromBytes(createImageWithoutData(asciiBytes(text)));
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals(CbfElementType.UNSIGNED_32_BIT_INTEGER, mdata.getDataElementType());
  }

  @Test
  public void testReadMetadataMimeHeader05() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    String text = createImageText().replaceFirst(
      "X-Binary-Element-Byte-Order: LITTLE_ENDIAN",
      "X-Binary-Element-Byte-Order: BIG_ENDIAN");
    ImageInputStream stream = streamFromBytes(createImageWithoutData(asciiBytes(text)));
    reader.setInput(stream);
    CbfMetadata mdata = (CbfMetadata)reader.getImageMetadata(0);
    assertEquals(ByteOrder.BIG_ENDIAN, mdata.getDataElementByteOrder());
  }

  /* Element type: unsigned 8-bit integer; compression: none. */
  @Test
  public void testRead01() throws IOException {
    CbfImageReaderSpi spi = new CbfImageReaderSpi();
    CbfImageReader reader = (CbfImageReader)spi.createReaderInstance();
    String text = createImageText().replaceFirst(
      "conversions=\"x-CBF_BYTE_OFFSET\"",
      "conversions=\"x-CBF_NONE\"");
    text = text.replaceFirst(
      "X-Binary-Size: 6224001",
      "X-Binary-Size: 4");
    text = text.replaceFirst(
      "X-Binary-Element-Type: \"signed 32-bit integer\"",
      "X-Binary-Element-Type: \"unsigned 8-bit integer\"");
    text = text.replaceFirst(
      "Content-MD5: Ti67wd6I6DPuGIsOnMalMw==\r\n",
      "");
    text = text.replaceFirst(
      "X-Binary-Number-of-Elements: 6224001",
      "X-Binary-Number-of-Elements: 4");
    text = text.replaceFirst(
      "X-Binary-Size-Fastest-Dimension: 2463",
      "X-Binary-Size-Fastest-Dimension: 2");
    text = text.replaceFirst(
      "X-Binary-Size-Second-Dimension: 2527",
      "X-Binary-Size-Second-Dimension: 2");
    byte[] imageWithoutData = createImageWithoutData(asciiBytes(text));
    byte[] imageData = { 1, 2, 3, 4 };
    byte[] imageWithData = new byte[imageWithoutData.length + imageData.length];
    System.arraycopy(imageWithoutData, 0, imageWithData, 0, imageWithoutData.length);
    System.arraycopy(imageData, 0, imageWithData, imageWithoutData.length, imageData.length);
    ImageInputStream stream = streamFromBytes(imageWithData);
    reader.setInput(stream);
    BufferedImage image = reader.read(0);
    Raster raster = image.getData();
    assertEquals(1, raster.getNumBands());
    assertEquals(8, image.getColorModel().getComponentSize(0));
    assertEquals(2, raster.getWidth());
    assertEquals(2, raster.getHeight());
    assertEquals(BufferedImage.TYPE_BYTE_GRAY, image.getType());
    assertEquals(imageData[0], (byte)raster.getSample(0, 0, 0));
    assertEquals(imageData[1], (byte)raster.getSample(1, 0, 0));
    assertEquals(imageData[2], (byte)raster.getSample(0, 1, 0));
    assertEquals(imageData[3], (byte)raster.getSample(1, 1, 0));
  }

  /* Element type: signed 32-bit integer; compression: byte_offset. */
  @Test
  public void testRead02() throws IOException {
    int[][] expected = {
      { +1, +2 },
      { -1, -2 }
    };
    BufferedImage outImage = createIntBufferedImage(expected);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /*
   * Element type: signed 32-bit integer; compression: byte_offset. Min and
   * max values.
   */
  @Test
  public void testRead03() throws IOException {
    int[][] expected = {
      { 0, Integer.MAX_VALUE },
      { 5, Integer.MIN_VALUE }
    };
    BufferedImage outImage = createIntBufferedImage(expected);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Not square dimensions. */
  @Test
  public void testRead04() throws IOException {
    int[][] expected = {
      { 0, 1, 2 },
      { 3, 4, 5 }
    };
    BufferedImage outImage = createIntBufferedImage(expected);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /*
   * Element type: signed 32-bit integer; compression: byte_offset. Short
   * size values.
   */
  @Test
  public void testRead05() throws IOException {
    int[][] expected = {
      { Short.MIN_VALUE + 1, Short.MAX_VALUE },
      { 5, 6 }
    };
    BufferedImage outImage = createIntBufferedImage(expected);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage));
    assertEquals(BufferedImage.TYPE_CUSTOM, inImage.getType());
    assertEquals(DataBuffer.TYPE_INT, inImage.getSampleModel().getDataType());
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region. */
  @Test
  public void testRead06() throws IOException {
    int[][] output = {
      { 1, 2, 3 },
      { 4, 5, 6 }
    };
    int[][] expected = {
      { 1, 2 },
      { 4, 5 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceRegion(new Rectangle(0, 0, 2, 2));
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region. */
  @Test
  public void testRead07() throws IOException {
    int[][] output = {
      { 1, 2, 3 },
      { 4, 5, 6 }
    };
    int[][] expected = {
      { 2, 3 },
      { 5, 6 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceRegion(new Rectangle(1, 0, 2, 2));
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region. */
  @Test
  public void testRead08() throws IOException {
    int[][] output = {
      { 1, 2, 3 },
      { 4, 5, 6 },
      { 7, 8, 9 }
    };
    int[][] expected = {
      { 4, 5 },
      { 7, 8 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceRegion(new Rectangle(0, 1, 2, 2));
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region. */
  @Test
  public void testRead09() throws IOException {
    int[][] output = {
      { 1, 2, 3 },
      { 4, 5, 6 },
      { 7, 8, 9 }
    };
    int[][] expected = {
      { 5, 6 },
      { 8, 9 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceRegion(new Rectangle(1, 1, 2, 2));
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region. */
  @Test
  public void testRead10() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13 },
      { 14, 15, 16, 17 },
      { 18, 19, 20, 21 },
      { 22, 23, 24, 25 }
    };
    int[][] expected = {
      { 15, 16 },
      { 19, 20 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceRegion(new Rectangle(1, 1, 2, 2));
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region. */
  @Test
  public void testRead11() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13 },
      { 14, 15, 16, 17 },
      { 18, 19, 20, 21 },
      { 22, 23, 24, 25 }
    };
    int[][] expected = {
      { 10, 11, 12, 13 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceRegion(new Rectangle(0, 0, 4, 1));
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region. */
  @Test
  public void testRead12() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13 },
      { 14, 15, 16, 17 },
      { 18, 19, 20, 21 },
      { 22, 23, 24, 25 }
    };
    int[][] expected = {
      { 15, 16, 17 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceRegion(new Rectangle(1, 1, 3, 1));
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region. */
  @Test
  public void testRead13() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13 },
      { 14, 15, 16, 17 },
      { 18, 19, 20, 21 },
      { 22, 23, 24, 25 }
    };
    int[][] expected = {
      { 10 },
      { 14 },
      { 18 },
      { 22 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceRegion(new Rectangle(0, 0, 1, 4));
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region. */
  @Test
  public void testRead14() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13 },
      { 14, 15, 16, 17 },
      { 18, 19, 20, 21 },
      { 22, 23, 24, 25 }
    };
    int[][] expected = {
      { 15 },
      { 19 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceRegion(new Rectangle(1, 1, 1, 2));
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region of size 1x1. */
  @Test
  public void testRead15() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13 },
      { 14, 15, 16, 17 },
      { 18, 19, 20, 21 },
      { 22, 23, 24, 25 }
    };
    int[][] expected = {
      { 25 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceRegion(new Rectangle(3, 3, 1, 1));
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* 2x2 subsampling without offset. */
  @Test
  public void testRead16() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13 },
      { 14, 15, 16, 17 },
      { 18, 19, 20, 21 },
      { 22, 23, 24, 25 }
    };
    int[][] expected = {
      { 10, 12 },
      { 18, 20 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceSubsampling(2, 2, 0, 0);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* 2x1 subsampling without offset. */
  @Test
  public void testRead17() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13 },
      { 14, 15, 16, 17 },
      { 18, 19, 20, 21 },
      { 22, 23, 24, 25 }
    };
    int[][] expected = {
      { 10, 12 },
      { 14, 16 },
      { 18, 20 },
      { 22, 24 },
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceSubsampling(2, 1, 0, 0);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* 1x2 subsampling without offset. */
  @Test
  public void testRead18() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13 },
      { 14, 15, 16, 17 },
      { 18, 19, 20, 21 },
      { 22, 23, 24, 25 }
    };
    int[][] expected = {
      { 10, 11, 12, 13 },
      { 18, 19, 20, 21 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceSubsampling(1, 2, 0, 0);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* 3x3 subsampling without offset. */
  @Test
  public void testRead19() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13 },
      { 14, 15, 16, 17 },
      { 18, 19, 20, 21 },
      { 22, 23, 24, 25 }
    };
    int[][] expected = {
      { 10, 13 },
      { 22, 25 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceSubsampling(3, 3, 0, 0);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* 2x2 subsampling with offset. */
  @Test
  public void testRead20() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13 },
      { 14, 15, 16, 17 },
      { 18, 19, 20, 21 },
      { 22, 23, 24, 25 }
    };
    int[][] expected = {
      { 15, 17 },
      { 23, 25 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceSubsampling(2, 2, 1, 1);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* 2x1 subsampling with X offset. */
  @Test
  public void testRead21() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13 },
      { 14, 15, 16, 17 },
      { 18, 19, 20, 21 },
      { 22, 23, 24, 25 }
    };
    int[][] expected = {
      { 11, 13 },
      { 15, 17 },
      { 19, 21 },
      { 23, 25 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceSubsampling(2, 1, 1, 0);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* 1x2 subsampling with Y offset. */
  @Test
  public void testRead22() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13 },
      { 14, 15, 16, 17 },
      { 18, 19, 20, 21 },
      { 22, 23, 24, 25 }
    };
    int[][] expected = {
      { 14, 15, 16, 17 },
      { 22, 23, 24, 25 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceSubsampling(1, 2, 0, 1);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* 3x2 subsampling with offset. */
  @Test
  public void testRead23() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 }
    };
    int[][] expected = {
      { 18, 21 },
      { 30, 33 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceSubsampling(3, 2, 2, 1);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region and 2x2 subsampling. */
  @Test
  public void testRead24() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = {
      { 17, 19 },
      { 29, 31 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceRegion(new Rectangle(1, 1, 4, 4));
    param.setSourceSubsampling(2, 2, 0, 0);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Source region and 2x2 subsampling with offset. */
  @Test
  public void testRead25() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    int[][] expected = {
      { 24, 26 },
      { 36, 38 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    ImageReadParam param = createImageReadParam();
    param.setSourceRegion(new Rectangle(1, 1, 4, 4));
    param.setSourceSubsampling(2, 2, 1, 1);
    BufferedImage inImage = readBufferedImage(writeBufferedImage(outImage), param);
    assertArrayEquals(expected, getIntPixels(inImage));
  }

  /* Progress. */
  @Test
  public void testRead26() throws IOException {
    int[][] output = {
      { 10, 11, 12, 13, 14, 15 },
      { 16, 17, 18, 19, 20, 21 },
      { 22, 23, 24, 25, 26, 27 },
      { 28, 29, 30, 31, 32, 33 },
      { 34, 35, 36, 37, 38, 39 }
    };
    BufferedImage outImage = createIntBufferedImage(output);
    final List<Boolean> started = Arrays.asList(new Boolean[] { Boolean.FALSE });
    final List<Float> percentDone = new ArrayList<Float>();
    final List<Boolean> completed = Arrays.asList(new Boolean[] { Boolean.FALSE });
    final List<Boolean> aborted = Arrays.asList(new Boolean[] { Boolean.FALSE });
    IIOReadProgressListener listener = new IIOReadProgressListener() {
      public void sequenceStarted(ImageReader source, int minIndex) {
      }

      public void sequenceComplete(ImageReader source) {
      }

      public void imageStarted(ImageReader source, int imageIndex) {
        started.set(0, Boolean.TRUE);
      }

      public void imageProgress(ImageReader source, float percentageDone) {
        percentDone.add(percentageDone);
      }

      public void imageComplete(ImageReader source) {
        completed.set(0, Boolean.TRUE);
      }

      public void thumbnailStarted(ImageReader source, int imageIndex, int thumbnailIndex) {
      }

      public void thumbnailProgress(ImageReader source, float percentageDone) {
      }

      public void thumbnailComplete(ImageReader source) {
      }

      public void readAborted(ImageReader source) {
        aborted.set(0, Boolean.TRUE);
      }
    };
    readBufferedImage(writeBufferedImage(outImage), listener);
    assertTrue(started.get(0));
    assertTrue(completed.get(0));
    assertFalse(aborted.get(0));
    assertEquals(5, percentDone.size());
    assertEquals(80.0f, percentDone.get(percentDone.size() - 1), 0.1f);
  }

  /*
   * Read into second band (index 1) of RGB destination image.
   */
  @Test
  public void testRead27() throws IOException {
    int[][][] output = {
      { { 0, 0, 0 }, { 0, +1, 0 }, { 0, +2, 0 } },
      { { 0, 3, 0 }, { 0, +4, 0 }, { 0, +5, 0 } },
      { { 0, 6, 0 }, { 0, +7, 0 }, { 0, +8, 0 } },
      { { 0, 9, 0 }, { 0, 10, 0 }, { 0, 11, 0 } }
    };
    int[][][] expected = output;

    BufferedImage outImage = createIntRgbBufferedImage(output);
    ImageWriteParam outParam = createImageWriteParam();
    outParam.setSourceBands(new int[] { 1 });
    CbfMetadata outMdata = new CbfMetadata();
    outMdata.setDataElementType(CbfElementType.SIGNED_32_BIT_INTEGER);
    byte[] outBytes = writeBufferedImage(outImage, outMdata, outParam);
    BufferedImage inImage = createIntRgbBufferedImage(expected[0].length, expected.length);
    ImageReadParam inParam = createImageReadParam();
    inParam.setDestination(inImage);
    inParam.setDestinationBands(new int[] { 1 });
    BufferedImage inImage2 = readBufferedImage(outBytes, inParam);
    assertSame(inImage, inImage2);
    assertArrayEquals(expected, getIntRgbPixels(inImage));
  }
}
